import copy
import enum
import json
import math
import random
from typing import Any, Dict


class ImageModel(enum.Enum):
    Anime_Curated = "safe-diffusion"
    Anime_Full = "nai-diffusion"
    Furry = "nai-diffusion-furry"


class ImageResolution(enum.Enum):
    Small_Portrait = (384, 640)
    Small_Landscape = (640, 384)
    Small_Square = (512, 512)

    Normal_Portrait = (512, 768)
    Normal_Landscape = (768, 512)
    Normal_Square = (640, 640)

    Large_Portrait = (512, 1024)
    Large_Landscape = (1024, 512)
    Large_Square = (1024, 1024)


class ImageSampler(enum.Enum):
    k_lms = "k_lms"
    k_euler = "k_euler"
    k_euler_ancestral = "k_euler_ancestral"
    plms = "plms"
    ddim = "ddim"


class UCPreset(enum.Enum):
    Preset_Low_Quality_Bad_Anatomy = 2
    Preset_Low_Quality = 1
    Preset_None = 0
    Preset_Custom = -1


class ImagePreset:
    _UC_Presets = {
        ImageModel.Anime_Curated: {
            UCPreset.Preset_Low_Quality_Bad_Anatomy: "nsfw, lowres, bad anatomy, bad hands, text, error, "
            "missing fingers, extra digit, fewer digits, cropped, worst quality, low quality, normal quality, "
            "jpeg artifacts, signature, watermark, twitter username, blurry",
            UCPreset.Preset_Low_Quality: "nsfw, lowres, text, cropped, worst quality, low quality, normal quality, "
            "jpeg artifacts, signature, watermark, twitter username, blurry",
            UCPreset.Preset_None: "lowres",
            UCPreset.Preset_Custom: "",
        },
        ImageModel.Anime_Full: {
            UCPreset.Preset_Low_Quality_Bad_Anatomy: "lowres, bad anatomy, bad hands, text, error, missing fingers, "
            "extra digit, fewer digits, cropped, worst quality, low quality, normal quality, jpeg artifacts, "
            "signature, watermark, username, blurry",
            UCPreset.Preset_Low_Quality: "lowres, text, cropped, worst quality, low quality, normal quality, "
            "jpeg artifacts, signature, watermark, username, blurry",
            UCPreset.Preset_None: "lowres",
            UCPreset.Preset_Custom: "",
        },
        ImageModel.Furry: {
            UCPreset.Preset_Low_Quality: "nsfw, worst quality, low quality, what has science done, what, "
            "nightmare fuel, eldritch horror, where is your god now, why",
            UCPreset.Preset_None: "low res",
            UCPreset.Preset_Custom: "",
        },
    }

    _TYPE_MAPPING = {
        "quality_toggle": bool,
        "resolution": (ImageResolution, tuple),
        "uc_preset": UCPreset,
        "n_samples": int,
        "seed": int,
        "sampler": ImageSampler,
        "noise": (int, float),
        "strength": (int, float),
        "scale": (int, float),
        "steps": int,
        "uc": str,
    }

    _DEFAULT = {
        "quality_toggle": True,
        "resolution": ImageResolution.Normal_Portrait,
        "uc_preset": UCPreset.Preset_Low_Quality_Bad_Anatomy,
        "n_samples": 1,
        "seed": 0,
        "sampler": ImageSampler.k_euler_ancestral,
        "noise": 0.2,
        "strength": 0.7,
        "steps": 28,
        "scale": 11,
        "uc": "",
    }

    _settings: Dict[str, Any]

    # Seed provided when generating an image with seed 0 (default). Seed is also in metadata, but might be a hassle
    last_seed: int

    def __init__(self, **kwargs):
        self._settings = self._DEFAULT.copy()
        self.update(kwargs)

        self.last_seed = 0

    def __setitem__(self, o: str, v: Any):
        assert o in self._TYPE_MAPPING, f"'{o}' is not a valid setting"
        assert isinstance(
            v, self._TYPE_MAPPING[o]
        ), f"Expected type '{self._TYPE_MAPPING[o]}' for {o}, but got type '{type(v)}'"

        self._settings[o] = v

    def update(self, values: Dict[str, Any]) -> "ImagePreset":
        for k, v in values.items():
            self[k] = v

        return self

    def copy(self) -> "ImagePreset":
        return ImagePreset(**self._settings)

    def __contains__(self, o: str):
        return o in self._settings

    def __getitem__(self, o: str):
        return self._settings[o]

    def to_settings(self, model: ImageModel) -> Dict[str, Any]:
        settings = copy.deepcopy(self._settings)

        resolution = settings.pop("resolution")
        if isinstance(resolution, ImageResolution):
            resolution = resolution.value
        settings["width"], settings["height"] = resolution

        if settings["seed"] == 0:
            settings["seed"] = random.randint(1, 0xFFFFFFFF)
            self.last_seed = settings["seed"]

        uc_preset: UCPreset = settings.pop("uc_preset")

        uc: str = settings.pop("uc")
        default_uc = self._UC_Presets[model][uc_preset]
        combined_uc = f"{default_uc}, {uc}" if uc else default_uc
        settings["uc"] = combined_uc

        sampler: ImageSampler = settings.pop("sampler")
        settings["sampler"] = sampler.value

        # special arguments kept for metadata purposes (no effect on result)
        settings["qualityToggle"] = settings.pop("quality_toggle")
        settings["ucPreset"] = uc_preset.value

        return settings

    def get_max_n_samples(self):
        resolution = self._settings["resolution"]

        if isinstance(resolution, ImageResolution):
            resolution = resolution.value

        w, h = resolution

        if w * h < 512 * 1024:
            return 4

        return 1

    def calculate_cost(self, is_opus: bool):
        steps = self._settings["steps"]
        n_samples = self._settings["n_samples"]
        resolution = self._settings["resolution"]

        if isinstance(resolution, ImageResolution):
            resolution = resolution.value

        w, h = resolution

        if is_opus and n_samples == 1 and steps <= 28 and w * h <= 640 * 640:
            return 0

        r = w * h / 1024 / 1024
        per_step = (15.266497014243718 * math.exp(r * 0.6326248927474729) - 15.225164493059737) / 28
        per_sample = max(math.ceil(per_step * steps), 2)

        return per_sample * n_samples

    @classmethod
    def from_file(cls, path: str) -> "ImagePreset":
        with open(path, encoding="utf-8") as f:
            data = json.loads(f.read())

        return cls(**data)

    def to_file(self, path: str):
        with open(path, "w", encoding="utf-8") as f:
            f.write(json.dumps(self._settings))



import base64
from hashlib import sha256
from typing import Any, AsyncIterable, Dict, Iterable, List, Optional, Tuple, Union

from novelai_api.BanList import BanList
from novelai_api.BiasGroup import BiasGroup
from novelai_api.GlobalSettings import GlobalSettings
from novelai_api.ImagePreset import ImageModel, ImagePreset
from novelai_api.Keystore import Keystore
from novelai_api.NovelAIError import NovelAIError
from novelai_api.Preset import Model, Preset
from novelai_api.utils import assert_type, compress_user_data, encrypt_user_data, get_access_key


class HighLevel:
    _parent: "NovelAIAPI"  # noqa: F821

    def __init__(self, parent: "NovelAIAPI"):  # noqa: F821
        self._parent = parent

    async def register(
        self,
        recapcha: str,
        email: str,
        password: str,
        send_mail: bool = True,
        giftkey: Optional[str] = None,
    ) -> bool:
        """
        Register a new account

        :param recapcha: Recapcha of the NovelAI website
        :param email: Email of the account (username)
        :param password: Password of the account
        :param send_mail: Send the mail (hashed and used for recovery)
        :param giftkey: Giftkey

        :return: True if success
        """

        assert assert_type(str, email=email)

        hashed_email = sha256(email.encode()).hexdigest() if send_mail else None
        key = get_access_key(email, password)
        return await self._parent.low_level.register(recapcha, key, hashed_email, giftkey)

    async def login(self, email: str, password: str) -> str:
        """
        Log in to the account

        :param email: Email of the account (username)
        :param password: Password of the account

        :return: User's access token
        """

        access_key = get_access_key(email, password)
        rsp = await self._parent.low_level.login(access_key)

        self._parent.headers["Authorization"] = f"Bearer {rsp['accessToken']}"

        return rsp["accessToken"]

    async def login_from_token(self, access_key: str):
        rsp = await self._parent.low_level.login(access_key)

        self._parent.headers["Authorization"] = f"Bearer {rsp['accessToken']}"

    async def get_keystore(self, key: bytes) -> Keystore:
        """
        Retrieve the keystore and decrypt it in a readable manner.
        The keystore is the mapping of meta -> encryption key of each object.
        If this function throws errors repeatedly at you,
        check your internet connection or the integrity of your keystore.
        Losing your keystore, or overwriting it means losing all content on the account.

        :param key: Account's encryption key

        :return: Keystore object
        """

        keystore = Keystore(await self._parent.low_level.get_keystore())
        keystore.decrypt(key)

        return keystore

    async def set_keystore(self, keystore: Keystore, key: bytes) -> bytes:
        keystore.encrypt(key)

        return await self._parent.low_level.set_keystore(keystore.data)

    async def download_user_stories(self) -> Dict[str, Dict[str, Union[str, int]]]:
        stories = await self._parent.low_level.download_objects("stories")

        return stories["objects"]

    async def download_user_story_contents(
        self,
    ) -> Dict[str, Dict[str, Union[str, int]]]:
        story_contents = await self._parent.low_level.download_objects("storycontent")

        return story_contents["objects"]

    async def download_user_presets(self) -> List[Dict[str, Union[str, int]]]:
        presets = await self._parent.low_level.download_objects("presets")

        return presets["objects"]

    async def download_user_modules(self) -> List[Dict[str, Union[str, int]]]:
        modules = await self._parent.low_level.download_objects("aimodules")

        return modules["objects"]

    async def download_user_shelves(self) -> List[Dict[str, Union[str, int]]]:
        modules = await self._parent.low_level.download_objects("shelf")

        return modules["objects"]

    async def upload_user_content(
        self,
        data: Dict[str, Any],
        encrypt: bool = False,
        keystore: Optional[Keystore] = None,
    ) -> bool:
        """
        Upload user content. If it has been decrypted with decrypt_user_data,
        it should be re-encrypted with encrypt_user_data, even if the decryption failed

        :param data: Object to upload
        :param encrypt: Encrypt/compress the data if True and not already encrypted
        :param keystore: Keystore to encrypt data if encrypt is True

        :return: True if the upload succeeded, False otherwise
        """

        object_id = data["id"]
        object_type = data["type"]
        object_meta = data["meta"]
        object_data = data["data"]

        if encrypt:
            if object_type in ("stories", "storycontent", "aimodules", "shelf"):
                assert keystore is not None, "Keystore is not set, cannot encrypt data"
                encrypt_user_data(data, keystore)
            elif object_type in ("presets",):
                compress_user_data(data)

        # clean data introduced by decrypt_user_data
        # this step should have been done in encrypt_user_data, but the user could have not called it
        for key in ("nonce", "compressed", "decrypted"):
            if key in object_data:
                self._parent.logger.warning(f"Data {key} left in object '{object_type}' of id '{object_id}'")
                del object_data[key]

        return await self._parent.low_level.upload_object(object_type, object_id, object_meta, object_data)

    async def upload_user_contents(self, datas: Iterable[Dict[str, Any]]) -> List[Tuple[str, Optional[NovelAIError]]]:
        """
        Upload multiple user contents. If the content has been decrypted with decrypt_user_data,
        it should be re-encrypted with encrypt_user_data, even if the decryption failed

        :param datas: Objects to upload

        :return: A list of (id, error) of all the objects that failed to be uploaded
        """

        status = []

        for data in datas:
            try:
                success = await self.upload_user_content(data)
                if not success:
                    status.append((data["id"], None))
            except NovelAIError as e:
                status.append((data["id"], e))

        return status

    async def _generate(
        self,
        prompt: Union[List[int], str],
        model: Model,
        preset: Preset,
        global_settings: GlobalSettings,
        bad_words: Optional[Union[Iterable[BanList], BanList]] = None,
        biases: Optional[Union[Iterable[BiasGroup], BiasGroup]] = None,
        prefix: Optional[str] = None,
        stream: bool = False,
        **kwargs,
    ):
        """
        Generate content from an AI on the NovelAI server which support streaming

        :param prompt: Context to give to the AI (raw text or list of tokens)
        :param model: Model to use for the AI
        :param preset: Preset to use for the generation settings
        :param global_settings: Global settings (used for generation)
        :param bad_words: Tokens to ban for this generation
        :param biases: Tokens to bias (up or down) for this generation
        :param prefix: Module to use for this generation
        :param stream: Use data streaming for the response

        :return: Content that has been generated
        """

        assert preset is not None, "Uninitialized preset"
        assert (
            preset.model == model
        ), f"Preset {preset.name} (model {preset.model}) is not compatible with model {model}"

        preset_params = preset.to_settings()
        global_params = global_settings.to_settings(model)

        params = {}

        params.update(preset_params)
        params.update(global_params)
        params.update(kwargs)

        params["prefix"] = "vanilla" if prefix is None else prefix

        if params["num_logprobs"] == GlobalSettings.NO_LOGPROBS:
            del params["num_logprobs"]

        if bad_words is not None:
            if isinstance(bad_words, BanList):
                bad_words = [bad_words]

            for i, bad_word in enumerate(bad_words):
                assert isinstance(
                    bad_word, BanList
                ), f"Expected type 'BanList' for item #{i} of bad_words, but got '{type(bad_word)}'"
                params["bad_words_ids"].extend(bad_word.get_tokenized_banlist(model))

        if biases is not None:
            if isinstance(biases, BiasGroup):
                biases = [biases]

            for i, bias in enumerate(biases):
                assert isinstance(
                    bias, BiasGroup
                ), f"Expected type 'BiasGroup' for item #{i} of biases, but got '{type(bias)}'"
                params["logit_bias_exp"].extend(bias.get_tokenized_biases(model))

        # Delete the options that return an unknown error (success status code, but server error)
        if "repetition_penalty_slope" in params and params["repetition_penalty_slope"] == 0:
            del params["repetition_penalty_slope"]

        if not params["bad_words_ids"]:
            del params["bad_words_ids"]

        async for i in self._parent.low_level.generate(prompt, model, params, stream):
            yield i

    async def generate(
        self,
        prompt: Union[List[int], str],
        model: Model,
        preset: Preset,
        global_settings: GlobalSettings,
        bad_words: Optional[Union[Iterable[BanList], BanList]] = None,
        biases: Optional[Union[Iterable[BiasGroup], BiasGroup]] = None,
        prefix: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Generate text from an AI on the NovelAI server which support streaming

        :param prompt: Context to give to the AI (raw text or list of tokens)
        :param model: Model to use for the AI
        :param preset: Preset to use for the generation settings
        :param global_settings: Global settings (used for generation)
        :param bad_words: Tokens to ban for this generation
        :param biases: Tokens to bias (up or down) for this generation
        :param prefix: Module to use for this generation

        :return: Content that has been generated
        """

        async for e in self._generate(
            prompt,
            model,
            preset,
            global_settings,
            bad_words,
            biases,
            prefix,
            False,
            **kwargs,
        ):
            return e

    async def generate_stream(
        self,
        prompt: Union[List[int], str],
        model: Model,
        preset: Preset,
        global_settings: GlobalSettings,
        bad_words: Optional[Union[Iterable[BanList], BanList]] = None,
        biases: Optional[Union[Iterable[BiasGroup], BiasGroup]] = None,
        prefix: Optional[str] = None,
        **kwargs,
    ) -> AsyncIterable[Dict[str, Any]]:
        """
        Generate text from an AI on the NovelAI server

        :param prompt: Context to give to the AI (raw text or list of tokens)
        :param model: Model to use for the AI
        :param preset: Preset to use for the generation settings
        :param global_settings: Global settings (used for generation)
        :param bad_words: Tokens to ban for this generation
        :param biases: Tokens to bias (up or down) for this generation
        :param prefix: Module to use for this generation

        :return: Content that has been generated
        """

        async for e in self._generate(
            prompt,
            model,
            preset,
            global_settings,
            bad_words,
            biases,
            prefix,
            True,
            **kwargs,
        ):
            yield e

    async def generate_image(
        self, prompt: str, model: ImageModel, preset: ImagePreset, **kwargs
    ) -> AsyncIterable[bytes]:
        """
        Generate image from an AI on the NovelAI server

        :param prompt: Prompt to give to the AI (raw text describing the wanted image)
        :param model: Model to use for the AI
        :param preset: Preset to use for the generation settings

        :return: Content that has been generated
        """

        settings = preset.to_settings(model)
        settings.update(kwargs)

        uc = settings["uc"]
        if "nsfw" in prompt and uc.startswith("nsfw,"):
            settings["uc"] = uc[len("nsfw,") :].strip()

        quality_toggle = preset["quality_toggle"]
        if quality_toggle:
            prompt = f"masterpiece, best quality, {prompt}"

        async for e in self._parent.low_level.generate_image(prompt, model, settings):
            yield base64.b64decode(e)



import json
import operator
from base64 import b64decode, b64encode, urlsafe_b64encode
from hashlib import blake2b
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union
from zlib import MAX_WBITS, Z_BEST_COMPRESSION
from zlib import compressobj as deflate_obj
from zlib import decompress as inflate

import argon2
from nacl.exceptions import CryptoError
from nacl.secret import SecretBox

from novelai_api.Keystore import Keystore
from novelai_api.Preset import Model, Preset
from novelai_api.Tokenizer import Tokenizer

# boilerplate
NoneType: type = type(None)


def assert_type(expected, **types):
    for k, v in types.items():
        assert isinstance(v, expected), f"Expected type '{expected}' for {k}, but got type '{type(v)}'"


operator_to_str = {
    operator.eq: "exactly {} characters",
    operator.lt: "less than {} characters",
    operator.le: "{} characters or less",
    operator.gt: "more than {} characters",
    operator.ge: "{} characters or more",
}


def assert_len(expected, op: operator = operator.eq, **values):
    op_str = operator_to_str[op].format(expected)

    for k, v in values.items():
        assert v is None or op(len(v), expected), f"'{k}' should be {op_str}, got length of {len(v)}'"


# API utils
def argon_hash(email: str, password: str, size: int, domain: str) -> str:
    pre_salt = f"{password[:6]}{email}{domain}"

    # salt
    blake = blake2b(digest_size=16)
    blake.update(pre_salt.encode())
    salt = blake.digest()

    raw = argon2.low_level.hash_secret_raw(
        password.encode(),
        salt,
        2,
        int(2000000 / 1024),
        1,
        size,
        argon2.low_level.Type.ID,
    )
    hashed = urlsafe_b64encode(raw).decode()

    return hashed


def get_access_key(email: str, password: str) -> str:
    assert_type(str, email=email, password=password)

    return argon_hash(email, password, 64, "novelai_data_access_key")[:64]


def get_encryption_key(email: str, password: str) -> bytes:
    assert_type(str, email=email, password=password)

    pre_key = argon_hash(email, password, 128, "novelai_data_encryption_key").replace("=", "")

    blake = blake2b(digest_size=32)
    blake.update(pre_key.encode())

    return blake.digest()


COMPRESSION_PREFIX = b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01"


def decrypt_data(
    data: Union[str, bytes], key: bytes, nonce: Optional[bytes] = None
) -> Union[Tuple[str, bytes, bool], Tuple[None, None, bool]]:
    box = SecretBox(key)

    if not isinstance(data, bytes):
        data = data.encode()

    # data is compressed
    is_compressed = data.startswith(COMPRESSION_PREFIX)
    if is_compressed:
        data = data[len(COMPRESSION_PREFIX) :]

    if nonce is None:
        nonce = data[: box.NONCE_SIZE]
        data = data[box.NONCE_SIZE :]

    try:
        data = box.decrypt(data, nonce=nonce)
        if is_compressed:
            data = inflate(data, -MAX_WBITS)

        return data.decode(), nonce, is_compressed
    except CryptoError:
        return None, None, False


def encrypt_data(
    data: Union[str, bytes],
    key: bytes,
    nonce: Optional[bytes] = None,
    is_compressed: bool = False,
) -> bytes:

    box = SecretBox(key)

    if not isinstance(data, bytes):
        data = data.encode()

    # NOTE: zlib results in different data than the library used by NAI, but they are fully compatible
    if is_compressed:
        deflater = deflate_obj(Z_BEST_COMPRESSION, wbits=-MAX_WBITS)
        data = deflater.compress(data) + deflater.flush()

    data = bytes(box.encrypt(data, nonce))

    if is_compressed:
        data = COMPRESSION_PREFIX + data

    return data


# function injection to avoid circular import
if not hasattr(Keystore, "_encrypt_data"):
    Keystore._encrypt_data = encrypt_data

if not hasattr(Keystore, "_decrypt_data"):
    Keystore._decrypt_data = decrypt_data


def decompress_user_data(items: Union[List[Dict[str, Any]], Dict[str, Any]]):
    """
    Decompress the data of each item in :ref: items
    Doesn't decrypt, but does a b64 to UTF8 translation
    """

    if not isinstance(items, (list, tuple)):
        items = [items]

    for item in items:
        assert isinstance(item, dict), f"Expected type 'dict' for item of 'items', got type '{type(item)}'"
        assert "data" in item, "Expected key 'data' in item"

        # skip already decompressed data
        if item.get("decrypted"):
            continue

        try:
            data = b64decode(item["data"])

            is_compressed = data.startswith(COMPRESSION_PREFIX)
            if is_compressed:
                data = data[len(COMPRESSION_PREFIX) :]
                data = inflate(data, -MAX_WBITS)

            item["data"] = json.loads(data.decode())
            item["decrypted"] = True  # not decrypted, per se, but for genericity
            item["compressed"] = is_compressed
        except json.JSONDecodeError:
            item["decrypted"] = False


def compress_user_data(items: Union[List[Dict[str, Any]], Dict[str, Any]]):
    """
    Compress the data of each item in :ref: items
    Doesn't encrypt, but does a UTF8 to b64 translation
    Must have been decompressed by decompress_user_data()
    """

    if not isinstance(items, (list, tuple)):
        items = [items]

    for item in items:
        assert isinstance(item, dict), f"Expected type 'dict' for item of 'items', got type '{type(item)}'"
        assert "data" in item, "Expected key 'data' in item"

        if "decrypted" in item:
            if item["decrypted"]:
                data = json.dumps(item["data"], separators=(",", ":"), ensure_ascii=False).encode()

                if "compressed" in item:
                    if item["compressed"]:
                        deflater = deflate_obj(Z_BEST_COMPRESSION, wbits=-MAX_WBITS)
                        data = deflater.compress(data) + deflater.flush()
                        data = COMPRESSION_PREFIX + data
                    del item["compressed"]

                item["data"] = b64encode(data).decode()
            del item["decrypted"]


def decrypt_user_data(items: Union[List[Dict[str, Any]], Dict[str, Any]], keystore: Keystore):
    """
    Decrypt the data of each item in :ref: items
    If an item has already been decrypted, it won't be decrypted a second time

    :param items: Item or list of items to decrypt
    :param keystore: Keystore retrieved with the get_keystore method
    """

    # 1 item
    if not isinstance(items, (list, tuple)):
        items = [items]

    for item in items:
        assert isinstance(item, dict), f"Expected type 'dict' for item of 'items', got type '{type(item)}'"

        if item.get("decrypted"):
            continue

        # FIXME: replace the assert by meaningful errors ? Expect the data to be right ?
        assert "data" in item, "Expected key 'data' in item"
        assert "meta" in item, "Expected key 'meta' in item"

        meta = item["meta"]
        #       assert meta in keystore
        if meta not in keystore:
            print("Meta missing:", meta)
        else:
            key = keystore[meta]

            data, nonce, is_compressed = decrypt_data(b64decode(item["data"]), key)
            if data is not None:
                try:
                    data = json.loads(data)
                    item["data"] = data
                    item["nonce"] = nonce
                    item["decrypted"] = True
                    item["compressed"] = is_compressed
                    continue

                except json.JSONDecodeError:
                    pass

        item["decrypted"] = False


def encrypt_user_data(items: Union[List[Dict[str, Any]], Dict[str, Any]], keystore: Keystore):
    """
    Encrypt the data of each item in :ref: items
    If an item has already been encrypted, it won't be encrypted a second time
    Must have been decrypted by decrypt_user_data()

    :param items: Item or list of items to encrypt
    :param keystore: Keystore retrieved with the get_keystore method
    """

    # 1 item
    if not isinstance(items, (list, tuple)):
        items = [items]

    for item in items:
        assert isinstance(item, dict), f"Expected type 'dict' for item of 'items', got type '{type(item)}'"

        if "decrypted" in item:
            if item["decrypted"]:
                # FIXME: replace the assert by meaningful errors ? Expect the data to be right ?
                assert "data" in item, "Expected key 'data' in item"
                assert "meta" in item, "Expected key 'meta' in item"
                assert "nonce" in item, "Expected key 'nonce' in item"
                assert "compressed" in item, "Expected key 'compressed' in item"

                meta = item["meta"]
                #            assert meta in keystore["keys"]
                if meta not in keystore:
                    print("Meta missing:", meta)
                else:
                    key = keystore[meta]

                    data = json.dumps(item["data"], separators=(",", ":"), ensure_ascii=False)
                    data = b64encode(encrypt_data(data, key, item["nonce"], item["compressed"])).decode()

                    item["data"] = data
                    del item["nonce"]
                    del item["compressed"]

            del item["decrypted"]


def assign_content_to_story(
    stories: Dict[str, Union[str, int, Dict[str, Any]]],
    story_contents: Union[List[Dict[str, Any]], Dict[str, Any]],
):

    if not isinstance(stories, (list, tuple)):
        stories = [stories]

    if not isinstance(story_contents, (list, tuple)):
        story_contents = [story_contents]

    story_contents = {content["id"]: content for content in story_contents}

    for story in stories:
        if story.get("decrypted"):
            remote_id = story["data"].get("remoteStoryId")

            if remote_id and remote_id in story_contents and story_contents[remote_id].get("decrypted"):
                story["content"] = story_contents[remote_id]


def remove_non_decrypted_user_data(items: List[Dict[str, Any]]):
    i = 0

    while i < len(items):
        if items[i].get("decrypted", False) is False:
            items.pop(i)
        else:
            i += 1


def tokens_to_b64(tokens: Iterable[int]) -> str:
    return b64encode(b"".join(t.to_bytes(2, "little") for t in tokens)).decode()


def b64_to_tokens(b64: str) -> List[int]:
    b = b64decode(b64)

    return [int.from_bytes(b[i : i + 2], "little") for i in range(0, len(b), 2)]


def extract_preset_data(presets: List[Dict[str, Any]]) -> Dict[str, Preset]:
    preset_list = {}
    for preset_data in presets:
        decompress_user_data(preset_data)
        preset_list[preset_data["id"]] = Preset.from_preset_data(preset_data["data"])

    return preset_list


def tokenize_if_not(model: Model, o: Union[str, List[int]]) -> List[int]:
    if isinstance(o, list):
        return o

    assert isinstance(o, str)
    return Tokenizer.encode(model, o)


# TODO: story tree builder



import enum
import json
import operator
from typing import Any, AsyncIterable, Dict, List, NoReturn, Optional, Tuple, Union
from urllib.parse import quote, urlencode

from aiohttp import ClientSession
from aiohttp.client_reqrep import ClientResponse

from novelai_api.ImagePreset import ImageModel
from novelai_api.NovelAIError import NovelAIError
from novelai_api.Preset import Model
from novelai_api.SchemaValidator import SchemaValidator
from novelai_api.Tokenizer import Tokenizer
from novelai_api.utils import NoneType, assert_len, assert_type, tokens_to_b64


# === INTERNALS === #
# === API === #
class LowLevel:
    _parent: "NovelAIAPI"  # noqa: F821
    _is_async: bool

    is_schema_validation_enabled: bool

    def __init__(self, parent: "NovelAIAPI"):  # noqa: F821
        self._parent = parent
        self.is_schema_validation_enabled = True

    @staticmethod
    def _treat_response_object(rsp: ClientResponse, content: Any, status: int) -> Any:
        # error is an unexpected fail and usually come with a success status
        if isinstance(content, dict) and "error" in content:
            raise NovelAIError(rsp.status, content["error"])

        # success
        if rsp.status == status:
            return content

        # not success, but valid response
        if isinstance(content, dict) and "message" in content:  # NovelAI REST API error
            raise NovelAIError(rsp.status, content["message"])

        # HTTPException error
        if hasattr(rsp, "reason"):
            raise NovelAIError(rsp.status, str(rsp.reason))

        raise NovelAIError(rsp.status, "Unknown error")

    def _treat_response_bool(self, rsp: ClientResponse, content: Any, status: int) -> bool:
        if rsp.status == status:
            return True

        self._treat_response_object(rsp, content, status)
        return False

    @staticmethod
    async def _treat_response(rsp: ClientResponse, data: Any) -> Any:
        if rsp.content_type == "application/json":
            return await data.json()

        if rsp.content_type in ("text/plain", "text/html"):
            return await data.text()

        if rsp.content_type in ("audio/mpeg", "audio/webm"):
            return await data.read()

        raise RuntimeError(f"Unsupported type: {rsp.content_type}")

    @staticmethod
    def _parse_stream_data(stream_content: str) -> Dict[str, Any]:
        stream_data = {}

        for line in stream_content.strip("\n").splitlines():
            if line == "":
                continue

            colon = line.find(":")
            # TODO: replace by a meaningful error
            if colon == -1:
                raise NovelAIError(0, f"Malformed data stream line: '{line}'")

            stream_data[line[:colon]] = line[colon + 1 :].strip()

        return stream_data

    async def _treat_response_stream(self, rsp: ClientResponse, data: bytes) -> Any:
        data = data.decode()

        if rsp.content_type == "text/event-stream":
            stream_data = self._parse_stream_data(data)

            # TODO: replace by a meaningful error
            assert "data" in stream_data
            data = stream_data["data"]

            if data.startswith("{") and data.endswith("}"):
                data = json.loads(data)

        return data

    async def _request(
        self, method: str, url: str, session: ClientSession, data: Union[Dict[str, Any], str], stream: bool
    ):

        kwargs = {
            "timeout": self._parent._timeout,  # noqa
            "cookies": self._parent.cookies,
            "headers": self._parent.headers,
            "json" if isinstance(data, dict) else "data": data,
        }

        if self._parent.proxy is not None:
            kwargs["proxy"] = self._parent.proxy

        if self._parent.proxy_auth is not None:
            kwargs["proxy_auth"] = self._parent.proxy_auth

        async with session.request(method, url, **kwargs) as rsp:
            if stream:
                content = b""

                async for chunk in rsp.content.iter_any():
                    # the event can be in the middle of a chunk... tfw...
                    if content and b"event" in chunk:
                        event_pos = chunk.find(b"event:")
                        content += chunk[:event_pos]

                        yield rsp, await self._treat_response_stream(rsp, content)
                        content = chunk[event_pos:]
                    else:
                        # TODO: Is there no way to check for malformed chunks here ? Massively sucks.
                        #       .iter_chunks() doesn't help either, as the request doesn't fit in an HTTP chunk
                        content += chunk

                yield rsp, await self._treat_response_stream(rsp, content)
            else:
                yield rsp, await self._treat_response(rsp, rsp)

    async def request_stream(
        self, method: str, endpoint: str, data: Optional[Union[Dict[str, Any], str]] = None, stream: bool = True
    ):
        """
        Send request with support for data streaming

        :param method: Method of the request (get, post, delete)
        :param endpoint: Endpoint of the request
        :param data: Data to pass to the method if needed
        :param stream: Use data streaming for the response
        """

        url = f"{self._parent.BASE_ADDRESS}{endpoint}"

        is_sync = self._parent._session is None  # noqa
        session = ClientSession() if is_sync else self._parent._session  # noqa

        try:
            async for i in self._request(method, url, session, data, stream):
                yield i
        except Exception as e:
            raise e
        finally:
            if is_sync:
                await session.close()

    async def request(
        self, method: str, endpoint: str, data: Optional[Union[Dict[str, Any], str]] = None
    ) -> Tuple[ClientResponse, Any]:
        """
        Send request

        :param method: Method of the request (get, post, delete)
        :param endpoint: Endpoint of the request
        :param data: Data to pass to the method if needed
        """

        async for i in self.request_stream(method, endpoint, data, False):
            return i

    async def is_reachable(self) -> bool:
        """
        Check if the NovelAI API is reachable

        :return: True if reachable, False if not
        """

        rsp, content = await self.request("get", "/")
        return self._treat_response_bool(rsp, content, 200)

    async def register(
        self, recapcha: str, access_key: str, email: Optional[str], giftkey: Optional[str] = None
    ) -> bool:
        """
        Register a new account

        :param recapcha: Recapcha of the NovelAI website
        :param access_key: Access key of the account
        :param email: Hashed email (used for recovery)
        :param giftkey: Giftkey

        :return: True if success
        """

        assert_type(str, recapcha=recapcha, access_key=access_key)
        assert_type((str, NoneType), email=email, giftkey=giftkey)
        assert_len(64, access_key=access_key)
        assert_len(64, email=email)

        data = {"recapcha": recapcha, "key": access_key}

        if email is not None:
            data["email"] = email
        if giftkey is not None:
            data["giftkey"] = giftkey

        rsp, content = await self.request("post", "/user/register", data)
        self._treat_response_object(rsp, content, 201)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_SuccessfulLoginResponse", content)

        return content

    async def login(self, access_key: str) -> Dict[str, str]:
        """
        Log in to the account

        :param access_key: Access key of the account

        :return: Response of the request
        """

        assert_type(str, access_key=access_key)
        assert_len(64, access_key=access_key)

        rsp, content = await self.request("post", "/user/login", {"key": access_key})
        self._treat_response_object(rsp, content, 201)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_SuccessfulLoginResponse", content)

        return content

    async def change_access_key(
        self, current_key: str, new_key: str, new_email: Optional[str] = None
    ) -> Dict[str, str]:
        assert_type(str, current_key=current_key, new_key=new_key)
        assert_type((str, NoneType), new_email=new_email)
        assert_len(64, current_key=current_key, new_key=new_key)

        data = {"currentAccessKey": current_key, "newAccessKey": new_key}

        if new_email is not None:
            data["newEmail"] = new_email

        rsp, content = await self.request("post", "/user/change-access-key", data)
        self._treat_response_object(rsp, content, 200)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_SuccessfulLoginResponse", content)

        return content

    async def send_email_verification(self, email: str) -> bool:
        assert_type(str, email=email)

        rsp, content = await self.request("post", "/user/resend-email-verification", {"email": email})
        return self._treat_response_bool(rsp, content, 200)

    async def verify_email(self, verification_token: str) -> bool:
        assert_type(str, verification_token=verification_token)
        assert_len(64, verification_token=verification_token)

        rsp, content = await self.request("post", "/user/verify-email", {"verificationToken": verification_token})
        return self._treat_response_bool(rsp, content, 200)

    async def get_information(self) -> Dict[str, Any]:
        rsp, content = await self.request("get", "/user/information")
        self._treat_response_object(rsp, content, 200)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_AccountInformationResponse", content)

        return content

    async def request_account_recovery(self, email: str) -> bool:
        assert_type(str, email=email)

        rsp, content = await self.request("post", "/user/recovery/request", {"email": email})
        return self._treat_response_bool(rsp, content, 202)

    async def recover_account(self, recovery_token: str, new_key: str, delete_content: bool = False) -> Dict[str, Any]:
        assert_type(str, recovery_token=recovery_token, new_key=new_key)
        assert_type(bool, delete_content=delete_content)
        assert_len(16, operator.ge, recovery_token=recovery_token)
        assert_len(64, new_key=new_key)

        rsp, content = await self.request(
            "post",
            "/user/recovery/recover",
            {
                "recoveryToken": recovery_token,
                "newAccessKey": new_key,
                "deleteContent": delete_content,
            },
        )
        self._treat_response_object(rsp, content, 201)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_SuccessfulLoginResponse", content)

        return content

    async def delete_account(self) -> bool:
        rsp, content = await self.request("post", "/user/delete", None)
        return self._treat_response_bool(rsp, content, 200)

    async def get_data(self) -> Dict[str, Any]:
        rsp, content = await self.request("get", "/user/data")
        self._treat_response_object(rsp, content, 200)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_AccountInformationResponse", content)

        return content

    async def get_priority(self) -> Dict[str, Any]:
        rsp, content = await self.request("get", "/user/priority")
        self._treat_response_object(rsp, content, 200)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_PriorityResponse", content)

        return content

    class SubscriptionTier(enum.IntEnum):
        PAPER = 0
        TABLET = 1
        SCROLL = 2
        OPUS = 3

    async def get_subscription(self) -> Dict[str, Any]:
        rsp, content = await self.request("get", "/user/subscription")
        self._treat_response_object(rsp, content, 200)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_SubscriptionResponse", content)

        return content

    async def get_keystore(self) -> Dict[str, str]:
        rsp, content = await self.request("get", "/user/keystore")
        self._treat_response_object(rsp, content, 200)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_GetKeystoreResponse", content)

        return content

    async def set_keystore(self, keystore: Dict[str, str]) -> bool:
        assert_type(dict, keystore=keystore)

        rsp, content = await self.request("put", "/user/keystore", keystore)
        return self._treat_response_object(rsp, content, 200)

    async def download_objects(self, object_type: str) -> Dict[str, List[Dict[str, Union[str, int]]]]:
        assert_type(str, object_type=object_type)

        rsp, content = await self.request("get", f"/user/objects/{object_type}")
        self._treat_response_object(rsp, content, 200)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_ObjectsResponse", content)

        return content

    async def upload_objects(self, object_type: str, meta: str, data: str) -> bool:
        assert_type(str, object_type=object_type, meta=meta, data=data)
        assert_len(128, operator.le, meta=meta)

        rsp, content = await self.request("put", f"/user/objects/{object_type}", {"meta": meta, "data": data})
        self._treat_response_object(rsp, content, 200)

        return content

    async def download_object(self, object_type: str, object_id: str) -> Dict[str, Union[str, int]]:
        assert_type(str, object_type=object_type, object_id=object_id)

        rsp, content = await self.request("get", f"/user/objects/{object_type}/{object_id}")
        self._treat_response_object(rsp, content, 200)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_userData", content)

        return content

    async def upload_object(self, object_type: str, object_id: str, meta: str, data: str) -> bool:
        assert_type(str, object_type=object_type, object_id=object_id, meta=meta, data=data)
        assert_len(128, operator.le, meta=meta)

        rsp, content = await self.request(
            "patch",
            f"/user/objects/{object_type}/{object_id}",
            {"meta": meta, "data": data},
        )
        self._treat_response_object(rsp, content, 200)

        return content

    async def delete_object(self, object_type: str, object_id: str) -> Dict[str, Union[str, int]]:
        assert_type(str, object_type=object_type, object_id=object_id)

        rsp, content = await self.request("delete", f"/user/objects/{object_type}/{object_id}")
        return self._treat_response_object(rsp, content, 200)

    async def get_settings(self) -> str:
        rsp, content = await self.request("get", "/user/clientsettings")
        return self._treat_response_object(rsp, content, 200)

    async def set_settings(self, value: str) -> bool:
        assert_type(str, value=value)

        rsp, content = await self.request("put", "/user/clientsettings", value)
        return self._treat_response_bool(rsp, content, 200)

    async def bind_subscription(self, payment_processor: str, subscription_id: str) -> bool:
        assert_type(str, payment_processor=payment_processor, subscription_id=subscription_id)

        rsp, content = await self.request(
            "post",
            "/user/subscription/bind",
            {"paymentProcessor": payment_processor, "subscriptionId": subscription_id},
        )
        return self._treat_response_bool(rsp, content, 201)

    async def change_subscription(self, new_plan: str) -> bool:
        assert_type(str, new_plan=new_plan)

        rsp, content = await self.request("post", "/user/subscription/change", {"newSubscriptionPlan": new_plan})
        return self._treat_response_bool(rsp, content, 200)

    async def generate(
        self,
        prompt: Union[List[int], str],
        model: Model,
        params: Dict[str, Any],
        stream: bool = False,
    ):
        """
        :param prompt: Input to be sent the AI
        :param model: Model of the AI
        :param params: Generation parameters
        :param stream: Use data streaming for the response

        :return: Generated output
        """

        assert_type((str, list), prompt=prompt)
        assert_type(Model, model=model)
        assert_type(dict, params=params)
        assert_type(bool, stream=stream)

        if isinstance(prompt, str):
            prompt = Tokenizer.encode(model, prompt)

        prompt = tokens_to_b64(prompt)
        args = {"input": prompt, "model": model.value, "parameters": params}

        endpoint = "/ai/generate-stream" if stream else "/ai/generate"

        async for rsp, content in self.request_stream("post", endpoint, args, stream):
            self._treat_response_object(rsp, content, 201)

            yield content

    async def classify(self) -> NoReturn:
        raise NotImplementedError("Function is not implemented yet")

    async def train_module(self, data: str, rate: int, steps: int, name: str, desc: str) -> Dict[str, Any]:
        """
        :param data: Dataset of the module, in one single string
        :param rate: Learning rate of the training
        :param steps: Number of steps to train the module for
        :param name: Name of the module
        :param desc: Description of the module

        :return: Status of the module being trained
        """

        assert_type(str, data=data, name=name, desc=desc)
        assert_type(int, rate=rate, steps=steps)

        rsp, content = await self.request(
            "post",
            "/ai/module/train",
            {
                "data": data,
                "lr": rate,
                "steps": steps,
                "name": name,
                "description": desc,
            },
        )
        self._treat_response_object(rsp, content, 201)

        # TODO: verify response ?

        return content

    async def get_trained_modules(self) -> List[Dict[str, Any]]:
        """
        :return: List of modules trained or in training
        """

        rsp, content = await self.request("get", "/ai/module/all")
        self._treat_response_object(rsp, content, 200)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_AiModuleDtos", content)

        return content

    async def get_trained_module(self, module_id: str) -> Dict[str, Any]:
        """
        :param module_id: Id of the module

        :return: Selected module, trained or in training
        """

        assert_type(str, module_id=module_id)

        rsp, content = await self.request("get", f"/ai/module/{module_id}")
        self._treat_response_object(rsp, content, 200)

        if self.is_schema_validation_enabled:
            SchemaValidator.validate("schema_AiModuleDto", content)

        return content

    async def delete_module(self, module_id: str) -> Dict[str, Any]:
        """
        Delete the module with id :ref: `module_id`

        :param module_id: Id of the module

        :return: Module that got deleted
        """

        assert_type(str, module_id=module_id)

        rsp, content = await self.request("delete", f"/ai/module/{module_id}")
        self._treat_response_object(rsp, content, 200)

        # TODO: verify response ?

        return content

    async def generate_voice(self, text: str, seed: str, voice: int, opus: bool, version: str) -> Dict[str, Any]:
        """
        Generate the Text-to-Speech of :ref: `text` using the given seed and voice

        :param text: Text to synthetize into voice
        :param seed: Person to use the voice of
        :param voice: Index of the voice to use
        :param opus: True for WebM format, False for mp3 format
        :param version: Version of the TTS

        :return: TTS audio data of the text
        """

        assert_type(str, text=text, seed=seed, version=version)
        assert_type(int, voice=voice)
        assert_type(bool, opus=opus)

        # urlencode keeps capitalization on bool =_=
        opus = "true" if opus else "false"
        query = urlencode(
            {
                "text": text,
                "seed": seed,
                "voice": voice,
                "opus": opus,
                "version": version,
            },
            quote_via=quote,
        )

        rsp, content = await self.request("get", f"/ai/generate-voice?{query}")
        self._treat_response_object(rsp, content, 200)

        return content

    async def suggest_tags(self, tag: str, model: ImageModel):
        assert_type(str, tag=tag)
        assert_type(ImageModel, model=model)

        query = urlencode(
            {
                "model": model,
                "prompt": tag,
            },
            quote_via=quote,
        )

        rsp, content = await self.request("get", f"/ai/generate-image/suggest-tags?{query}")
        self._treat_response_object(rsp, content, 200)

        return content

    async def generate_image(self, prompt: str, model: ImageModel, parameters: Dict[str, Any]) -> AsyncIterable[str]:
        assert_type(str, prompt=prompt)
        assert_type(ImageModel, model=model)
        assert_type(dict, parameters=parameters)

        args = {
            "input": prompt,
            "model": model.value,
            "parameters": parameters,
        }

        async for rsp, content in self.request_stream("post", "/ai/generate-image", args):
            self._treat_response_object(rsp, content, 201)

            yield content



from copy import deepcopy
from json import dumps, loads
from time import time
from typing import Any, Dict, Iterable, Iterator, List, NoReturn, Optional, Union

from novelai_api import NovelAI_API
from novelai_api.BanList import BanList
from novelai_api.BiasGroup import BiasGroup
from novelai_api.GlobalSettings import GlobalSettings
from novelai_api.Idstore import Idstore
from novelai_api.Keystore import Keystore
from novelai_api.Preset import Model, Preset
from novelai_api.Tokenizer import Tokenizer
from novelai_api.utils import b64_to_tokens, decrypt_user_data, encrypt_user_data


def _get_time() -> int:
    """
    Get the current time, as formatted for createdAt and lastUpdatedAt

    :return: Current time with millisecond precision
    """

    return int(time() * 1000)


def _get_short_time() -> int:
    """
    Because some lastUpdatedAt only are precise to the second

    :return: Current time with second precision
    """

    return int(time())


def _set_nested_item(item: Dict[str, Any], val: Any, path: str):
    path = path.split(".")

    for key in path[:-1]:
        item = item[key]

    item[path[-1]] = val


class NovelAI_StoryProxy:
    TEXT_GENERATION_SETTINGS_VERSION = 2

    DEFAULT_MODEL = Model.Euterpe

    _parent: "NovelAI_Story"

    _api: NovelAI_API
    _key: bytes
    _story: Dict[str, Any]
    _storycontent: Dict[str, Any]
    _tree: List[int]

    banlists: List[BanList]
    biases: List[BiasGroup]
    model: Model
    preset: Preset
    prefix: str
    context_size: int

    def _handle_banlist(self, data: Dict[str, Any]) -> NoReturn:
        if "bannedSequenceGroups" not in data:
            data["bannedSequenceGroups"] = []

        ban_seq = data["bannedSequenceGroups"]
        self.banlists = [BanList(*seq["sequences"], enabled=seq["enabled"]) for seq in ban_seq]

    def _handle_biasgroups(self, data: Dict[str, Any]) -> NoReturn:
        if "phraseBiasGroup" not in data:
            data["phraseBiasGroups"] = []

        self.biases = []
        for bias in data["phraseBiasGroups"]:
            self.biases.append(BiasGroup.from_data(bias))

    def _handle_preset(self, data: Dict[str, Any]) -> NoReturn:
        settings = data["settings"]

        if "textGenerationSettingsVersion" not in settings:
            settings["textGenerationSettingsVersion"] = self.TEXT_GENERATION_SETTINGS_VERSION

        if "prefix" not in settings:
            settings["prefix"] = "vanilla"
        self.prefix = settings["prefix"]

        if "model" not in settings:
            settings["model"] = self.DEFAULT_MODEL.value
        self.model = Model(settings["model"])

        if "preset" not in settings:
            settings["preset"] = ""

        parameters = settings["parameters"]

        if "bad_words_ids" in parameters:
            self.banlists.append(BanList(*parameters["bad_words_ids"]))
            del parameters["bad_words_ids"]

        if "logit_bias_groups" in parameters:
            for bias in parameters["logit_bias_groups"]:
                self.biases.append(BiasGroup.from_data(bias))
            del parameters["logit_bias_groups"]

        self.preset = Preset.from_preset_data(settings)
        self.preset.name = settings["preset"]
        self.preset.model = self.model

    def __init__(
        self,
        parent: "NovelAI_Story",
        key: bytes,
        story: Dict[str, Any],
        storycontent: Dict[str, Any],
    ):
        self._parent = parent

        self._api = parent._api
        self._key = key
        self._story = story
        self._storycontent = storycontent
        self._tree = []

        data = storycontent["data"]

        print(dumps(data, indent=4))
        self._handle_banlist(data)
        self._handle_biasgroups(data)
        self._handle_preset(data)

        # FIXME: variable context size ? From global settings ?
        self.context_size = 2048

        # TODO: trimResponses
        # TODO: banBrackets
        # TODO: dynamicPenaltyRange

        # TODO: remember
        # TODO: AN

        # TODO: Lorebook

    def _create_datablock(self, fragment: Dict[str, str], end_offset: int):
        story = self._storycontent["data"]["story"]
        blocks = story["datablocks"]
        fragments = story["fragments"]

        cur_index = story["currentBlock"]
        cur_block = blocks[cur_index]

        story["step"] += 1

        frag_index = len(fragments)
        fragments.append(fragment)

        start = cur_block["endIndex"] + len(cur_block["dataFragment"]["data"])

        block = {
            "nextBlock": [],
            "prevBlock": cur_index,
            "origin": fragment["origin"],
            "startIndex": start,
            "endIndex": start + end_offset,
            "dataFragment": fragment,
            "fragmentIndex": frag_index,
            "removedFragments": [],
            "chain": False,
        }
        new_index = len(blocks)
        blocks.append(block)

        cur_block["nextBlock"].append(new_index)

        story["currentBlock"] = new_index
        self._tree.append(new_index)

    def __str__(self) -> str:
        story_fragments = self._storycontent["data"]["story"]["fragments"]

        story_content = "".join(fragment["data"] for fragment in story_fragments)

        # FIXME: handle edit

        return story_content

    def build_context(self) -> List[int]:
        tokens = []

        # TODO: Remember tokens
        # TODO: AN tokens

        # TODO: optimize for large stories ?
        # edit is a pain for input in token form, so we use it's string representation instead
        story_content = str(self)
        story_content_size = self.context_size

        # TODO: add option to remove superfluous spaces at the end

        # only tokenize the tail to handle large stories
        story_tokens = []
        while len(story_tokens) < self.context_size:
            story_content_size *= 2
            story_tokens = Tokenizer.encode(self.model, story_content[-story_content_size:])

            # whole story content is tokenized
            if len(story_content) < story_content_size:
                break

        story_tokens = story_tokens[-self.context_size :]

        # TODO: LB tokens

        # TODO: Order and cut everything to fit

        tokens.extend(story_tokens)

        # Internal assert, should never happen
        assert len(tokens) <= self.context_size

        return tokens

    async def generate(self) -> "NovelAI_StoryProxy":
        input = self.build_context()
        # FIXME: find why the output is garbage
        rsp = await self._api.high_level.generate(
            input,
            self.model,
            self.preset,
            self._parent.global_settings,
            self.banlists,
            self.biases,
            self.prefix,
        )

        output = Tokenizer.decode(self.model, b64_to_tokens(rsp["output"]))
        fragment = {"data": output, "origin": "ai"}

        self._create_datablock(fragment, 0)

    async def edit(self, start: int, end: int, replace: str):
        # FIXME: redo edit implementation

        fragment = {"data": replace, "origin": "edit"}

        self._create_datablock(fragment, end - start)

    async def undo(self):
        story = self._storycontent["data"]["story"]

        cur_index = story["currentBlock"]
        blocks = story["datablocks"]

        cur_block = blocks[cur_index]
        story["currentBlock"] = cur_block["prevBlock"]

    async def redo(self):
        story = self._storycontent["data"]["story"]

        cur_index = story["currentBlock"]
        blocks = story["datablocks"]

        cur_block = blocks[cur_index]
        story["currentBlock"] = cur_block["nextBlock"][-1]

    async def save(self, upload: bool = False) -> bool:
        encrypted_story = encrypt_user_data(deepcopy(self._story), self._parent._keystore)
        encrypted_storycontent = encrypt_user_data(deepcopy(self._storycontent), self._parent._keystore)

        success = True

        # TODO: keep local copy if upload ?
        if upload:
            success = success and await self._api.high_level.upload_user_content(encrypted_storycontent)
            success = success and await self._api.high_level.upload_user_content(encrypted_story)

        return success

    async def choose(self, index: int) -> NoReturn:
        story = self._storycontent["data"]["story"]

        cur_index = story["currentBlock"]
        blocks = story["datablocks"]

        cur_block = blocks[cur_index]
        next_blocks = cur_block["nextBlock"]
        assert 0 <= index < len(next_blocks), f"Expected index between 0 and {len(next_blocks)}, but got {index}"

        story["currentBlock"] = next_blocks[index]

    async def flatten(self) -> NoReturn:
        story = self._storycontent["data"]["story"]

        blocks = story["datablocks"]
        new_datablocks = [blocks[i] for i in self._tree]
        self._tree = [i for i in range(len(new_datablocks))]
        story["datablocks"] = new_datablocks

    async def delete(self):
        pass

    async def get_current_tree(self) -> List[Dict[str, Any]]:
        story = self._storycontent["data"]["story"]

        blocks = story["datablocks"]
        return [blocks[i] for i in self._tree]


class NovelAI_Story:
    _story_instances: Dict[str, NovelAI_StoryProxy]

    _api: NovelAI_API
    _keystore: Keystore
    _idstore: Idstore

    global_settings: GlobalSettings

    def __init__(self, api: NovelAI_API, keystore: Keystore, global_settings: GlobalSettings):
        self._api = api
        self._keystore = keystore
        self._idstore = Idstore()

        self.global_settings = global_settings

        self._story_instances = {}

    def __iter__(self) -> Iterator[NovelAI_StoryProxy]:
        return self._story_instances.__iter__()

    def __getitem__(self, story_id: str) -> NovelAI_StoryProxy:
        return self._story_instances[story_id]

    def __len__(self) -> int:
        return len(self._story_instances)

    def load(self, story: Dict[str, Any], storycontent: Dict[str, Any]) -> NovelAI_StoryProxy:
        """
        Load a story proxy from a story and storycontent object
        """
        story_meta = story["meta"]
        story_id = story["data"]["remoteStoryId"]

        assert (
            story_meta == storycontent["meta"]
        ), f"Expected meta {story_meta} for storycontent, but got meta {storycontent['meta']}"
        assert story_id == storycontent["id"], f"Missmached id: expected {story_id}, but got {storycontent['id']}"

        proxy = NovelAI_StoryProxy(self, self._keystore[story_meta], story, storycontent)

        # FIXME: ignore or overwrite if id exists ?
        self._story_instances[story_id] = proxy

        return proxy

    def loads(
        self, stories: Iterable[Dict[str, Any]], storycontents: Iterable[Dict[str, Any]]
    ) -> List[NovelAI_StoryProxy]:
        mapping = {}
        for story in stories:
            if story.get("decrypted"):
                mapping[story["data"]["remoteStoryId"]] = story

        loaded = []
        for storycontent in storycontents:
            if storycontent.get("decrypted"):
                story_id = storycontent["id"]

                if story_id not in mapping:
                    self._api._logger.warning(f"Storycontent {story_id} has no associated story")
                else:
                    proxy = self.load(mapping[story_id], storycontent)
                    del mapping[story_id]

                    loaded.append(proxy)

        for story_id in mapping.keys():
            self._api._logger.warning(f"Story {story_id} has no associated storycontent")

        return loaded

    async def load_from_remote(self) -> List[NovelAI_StoryProxy]:
        stories = await self._api.high_level.download_user_stories()
        storycontents = await self._api.high_level.download_user_story_contents()

        decrypt_user_data(stories, self._keystore)
        decrypt_user_data(storycontents, self._keystore)

        return self.loads(stories, storycontents)

    def create(self) -> NovelAI_StoryProxy:
        meta = self._keystore.create()
        current_time = _get_time()
        current_time_short = _get_short_time()

        with open("templates/template_empty_story.txt") as f:
            story = loads(f.read())

        # local overwrites
        id_story = self._idstore.create()
        for path, val in (
            ("id", id_story),
            ("meta", meta),
            ("data.id", meta),
            ("data.remoteStoryId", id_story),
            ("data.createdAt", current_time),
            ("data.lastUpdatedAt", current_time),
            ("lastUpdatedAt", current_time_short),
        ):
            _set_nested_item(story, val, path)

        with open("templates/template_empty_storycontent.txt") as f:
            storycontent = loads(f.read())

        # local overwrites
        id_storycontent = self._idstore.create()
        id_lore_default = ""  # FIXME: get id

        for path, val in (
            ("id", id_storycontent),
            ("meta", meta),
            ("lastUpdatedAt", current_time_short),
            ("data.contextDefaults.loreDefaults.id", id_lore_default),
            ("data.contextDefaults.loreDefaults.lastUpdatedAt", current_time),
        ):
            _set_nested_item(storycontent, val, path)

        proxy = self.load(story, storycontent)

        return proxy

    def select(self, story_id: str) -> Optional[NovelAI_StoryProxy]:
        """
        Select a story proxy from the previously created/loaded ones

        :param story_id: Id of the selected story

        :return: Story or None if the story does't exist in the handler
        """

        if story_id not in self._story_instances:
            return None

        return self._story_instances[story_id]

    def unload(self, story_id: str) -> NoReturn:
        """
        Unload a previously created/loaded story, free'ing the NovelAI_StoryProxy object
        """

        if story_id in self._story_instances:
            del self._story_instances[story_id]



from copy import deepcopy
from enum import Enum, IntEnum
from json import loads
from os import listdir
from os.path import abspath, dirname, exists, join
from random import choice
from typing import Any, Dict, List, NoReturn, Optional, Union


# NOTE: the noqa are there because of Enum, because Enum's type inference sucks
class Order(IntEnum):
    Temperature = 0
    Top_K = 1
    Top_P = 2
    TFS = 3
    Top_A = 4
    Typical_P = 5


NAME_TO_ORDER = {
    "tfs": Order.TFS,
    "temperature": Order.Temperature,
    "top_p": Order.Top_P,
    "top_k": Order.Top_K,
    "top_a": Order.Top_A,
    "typical_p": Order.Typical_P,
}

ORDER_TO_NAME = {
    Order.TFS: "tfs",
    Order.Temperature: "temperature",
    Order.Top_P: "top_p",
    Order.Top_K: "top_k",
    Order.Top_A: "top_a",
    Order.Typical_P: "typical_p",
}


def enum_contains(enum_class: Enum, value) -> bool:
    if not hasattr(enum_class, "enum_member_values"):
        enum_class.enum_member_values = list(e.value for e in enum_class)  # noqa

    values = enum_class.enum_member_values
    assert len(values), f"Empty enum class {enum_class}"

    return value in values


class StrEnum(str, Enum):
    pass


class Model(StrEnum):
    # Calliope = "2.7B"
    Sigurd = "6B-v4"
    Euterpe = "euterpe-v2"
    Krake = "krake-v2"

    Genji = "genji-jp-6b-v2"
    Snek = "genji-python-6b"

    HypeBot = "hypebot"
    Inline = "infillmodel"


class PresetView:
    model: Model
    _official_values: Dict[str, List["Preset"]]

    def __init__(self, model: Model, officials_values: Dict[str, List["Preset"]]):
        self.model = model
        self._officials_values = officials_values

    def __iter__(self):
        return self._officials_values[self.model.value].__iter__()


class _PresetMetaclass(type):
    _officials_values: Dict[str, List["Preset"]]

    def __getitem__(cls, model: Model):
        assert isinstance(model, Model)

        return PresetView(model, cls._officials_values)


class Preset(metaclass=_PresetMetaclass):
    # TODO
    # do_sample                     boolean
    # early_stopping                boolean
    # num_beams                     number
    # pad_token_id                  number
    # bos_token_id                  number
    # eos_token_id                  number
    # no_repeat_ngram_size          number
    # encoder_no_repeat_ngram_size	number
    # num_return_sequences          number
    # max_time                      number
    # num_beam_groups               number
    # get_hidden_states             boolean
    # next_word                     boolean
    # output_nonzero_probs          boolean
    # generate_until_sentence       boolean

    _TYPE_MAPPING = {
        "textGenerationSettingsVersion": int,
        "temperature": (int, float),
        "max_length": int,
        "min_length": int,
        "top_k": int,
        "top_a": (int, float),
        "top_p": (int, float),
        "typical_p": (int, float),
        "tail_free_sampling": (int, float),
        "repetition_penalty": (int, float),
        "repetition_penalty_range": int,
        "repetition_penalty_slope": (int, float),
        "repetition_penalty_frequency": (int, float),
        "repetition_penalty_presence": int,
        "repetition_penalty_whitelist": list,
        "length_penalty": (int, float),
        "diversity_penalty": (int, float),
        "order": list,
        "eos_token_id": int,
        "stop_sequences": list,
    }

    _officials: Dict[str, Dict[str, "Preset"]]
    _officials_values: Dict[str, List["Preset"]]
    _defaults: Dict[str, str]

    _enabled: List[bool]

    _settings: Dict[str, Any]
    name: str
    model: Model

    def __init__(self, name: str, model: Model, settings: Optional[Dict[str, Any]] = None):
        self.name = name
        self.model = model

        self._enabled = [True] * len(Order)

        self._settings = {}
        self.update(settings)

    def __setitem__(self, o: str, v: Any):
        assert o in self._TYPE_MAPPING, f"'{o}' is not a valid setting"
        assert isinstance(
            v, self._TYPE_MAPPING[o]  # noqa
        ), f"Expected type '{self._TYPE_MAPPING[o]}' for {o}, but got type '{type(v)}'"

        if o == "order":
            assert isinstance(v, list), f"Expected type 'List[int|Order] for order, but got type '{type(v)}'"

            for i, e in enumerate(v):
                assert isinstance(
                    e, (int, Order)
                ), f"Expected type 'int' or 'Order for order #{i}, but got type '{type(v[i])}'"

                if isinstance(e, int):
                    v[i] = Order(e)

        self._settings[o] = v

    def __contains__(self, o: str) -> bool:
        return o in self._settings

    def __getitem__(self, o: str) -> Optional[Any]:
        return self._settings.get(o)

    def __repr__(self) -> str:
        model = self.model.value if self.model is not None else "<?>"
        return f"Preset: '{self.name} ({model})'"

    def enable(self, **kwargs) -> "Preset":
        for o in Order:
            name = ORDER_TO_NAME[o]
            enabled = kwargs.pop(name, False)
            self._enabled[o.value] = enabled  # noqa

        assert len(kwargs) == 0, f"Invalid order name: {', '.join(kwargs)}"

        return self

    def to_settings(self) -> Dict[str, Any]:
        settings = deepcopy(self._settings)

        if "textGenerationSettingsVersion" in settings:
            del settings["textGenerationSettingsVersion"]  # not API relevant

        for i, o in enumerate(Order):
            if not self._enabled[i]:
                settings["order"].remove(o)

        return settings

    def to_file(self, path: str) -> NoReturn:
        raise NotImplementedError()

    def copy(self) -> "Preset":
        return Preset(self.name, self.model, deepcopy(self._settings))

    def set(self, name: str, value: Any) -> "Preset":
        self[name] = value

        return self

    def update(self, values: Dict[str, Any]) -> "Preset":
        for k, v in values.items():
            self[k] = v

        return self

    @classmethod
    def from_preset_data(cls, data: Dict[str, Any]) -> "Preset":
        name = data["name"] if "name" in data else "<?>"

        model_name = data["model"] if "model" in data else ""
        model = Model(model_name) if enum_contains(Model, model_name) else None  # noqa

        settings = data["parameters"] if "parameters" in data else {}

        order = settings["order"] if "order" in settings else {}
        settings["order"] = list(NAME_TO_ORDER[o["id"]] for o in order)

        # TODO: add support for token banning and bias in preset
        settings.pop("bad_words_ids", None)  # get rid of unsupported option
        settings.pop("logit_bias_exp", None)  # get rid of unsupported option
        settings.pop("logit_bias_groups", None)  # get rid of unsupported option

        c = cls(name, model, settings)

        enabled = {o["id"]: o["enabled"] for o in order}
        c.enable(**enabled)

        return c

    @classmethod
    def from_file(cls, path: str) -> "Preset":
        with open(path, encoding="utf-8") as f:
            data = loads(f.read())

        return cls.from_preset_data(data)

    @classmethod
    def from_official(cls, model: Model, name: Optional[str] = None) -> Union["Preset", None]:
        model_value: str = model.value  # noqa

        if name is None:
            preset = choice(cls._officials_values[model_value])
        else:
            preset = cls._officials[model_value].get(name)

        if preset is not None:
            preset = deepcopy(preset)

        return preset

    @classmethod
    def from_default(cls, model: Model) -> Union["Preset", None]:
        model_value: str = model.value  # noqa

        default = cls._defaults.get(model_value)
        if default is None:
            return None

        preset = cls._officials[model_value].get(default)
        if preset is None:
            return None

        return preset.copy()


def import_officials():
    cls = Preset

    cls._officials_values = {}
    cls._officials = {}
    cls._defaults = {}

    for model in Model:
        model: Model

        path = join(
            dirname(abspath(__file__)),
            "presets",
            f"presets_{model.value.replace('-', '_')}",
        )

        if exists(join(path, "default.txt")):
            with open(join(path, "default.txt"), encoding="utf-8") as f:
                cls._defaults[model.value] = f.read().splitlines()[0]

        officials = {}
        for filename in listdir(path):
            if filename.endswith(".preset"):
                preset = cls.from_file(join(path, filename))
                officials[preset.name] = preset

        cls._officials_values[model.value] = list(officials.values())
        cls._officials[model.value] = officials


if not hasattr(Preset, "_officials"):
    import_officials()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['novelai_api', 'novelai_api.tokenizers']

package_data = \
{'': ['*'],
 'novelai_api': ['presets/presets_2.7B/*',
                 'presets/presets_6B_v4/*',
                 'presets/presets_euterpe_v2/*',
                 'presets/presets_genji_jp_6b_v2/*',
                 'presets/presets_genji_python_6b/*',
                 'presets/presets_hypebot/*',
                 'presets/presets_infillmodel/*',
                 'presets/presets_krake_v2/*',
                 'schemas/*',
                 'templates/*']}

install_requires = \
['PyNaCl>=1.5.0,<2.0.0',
 'aiohttp[speedups]>=3.8.3,<4.0.0',
 'argon2-cffi>=21.3.0,<22.0.0',
 'ftfy>=6.1.1,<7.0.0',
 'jsonschema>=4.17.0,<5.0.0',
 'regex>=2022.10.31,<2023.0.0',
 'tokenizers>=0.13.1,<0.14.0']

setup_kwargs = {
    'name': 'novelai-api',
    'version': '0.10.4',
    'description': 'Python API for the NovelAI REST API',
    'long_description': '# novelai-api\nPython API for the NovelAI REST API\n\nThis module is intended to be used by developers as a helper for using NovelAI\'s REST API.\n\n[TODO]: # (Add Quality Checking workflows and badges)\n\n| Category         | Badges                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        |\n|------------------|-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|\n| Pypi             | [![PyPI](https://img.shields.io/pypi/v/novelai-api)](https://pypi.org/project/novelai-api) [![PyPI - Python Version](https://img.shields.io/pypi/pyversions/novelai-api)](https://pypi.org/project/novelai-api) [![PyPI - License](https://img.shields.io/pypi/l/novelai-api)](https://pypi.org/project/novelai-api/) [![PyPI - Format](https://img.shields.io/pypi/format/novelai-api)](https://pypi.org/project/novelai-api/)                                                                                                                                                                                                                                                                                               |\n| Quality checking | [![Python package](https://github.com/Aedial/novelai-api/actions/workflows/python-package.yml/badge.svg)](https://github.com/Aedial/novelai-api/actions/workflows/python-package.yml) [![Python package](https://github.com/Aedial/novelai-api/actions/workflows/codeql-analysis.yml/badge.svg)](https://github.com/Aedial/novelai-api/actions/workflows/codeql-analysis.yml) [![linting: pylint](https://img.shields.io/badge/linting-pylint-yellowgreen)](https://github.com/PyCQA/pylint) [![security: bandit](https://img.shields.io/badge/security-bandit-yellow.svg)](https://github.com/PyCQA/bandit) [![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black) |\n| Stats            | [![GitHub top language](https://img.shields.io/github/languages/top/Aedial/novelai-api)](https://github.com/Aedial/novelai-api/search?l=python) ![Libraries.io dependency status for GitHub repo](https://img.shields.io/librariesio/github/Aedial/novelai-api) ![GitHub repo size](https://img.shields.io/github/repo-size/Aedial/novelai-api) ![GitHub issues](https://img.shields.io/github/issues-raw/Aedial/novelai-api) ![GitHub pull requests](https://img.shields.io/github/issues-pr-raw/Aedial/novelai-api)                                                                                                                                                                                                         |\n| Activity         | ![GitHub last commit](https://img.shields.io/github/last-commit/Aedial/novelai-api) ![GitHub commits since tagged version](https://img.shields.io/github/commits-since/Aedial/novelai-api/v0.10.3) ![GitHub commit activity](https://img.shields.io/github/commit-activity/m/Aedial/novelai-api)                                                                                                                                                                                                                                                                                                                                                                                                                              |\n\n\n### Prerequisites\nBefore anything, ensure that nox is installed (pip install nox).\nFor logging in, credentials are needed (NAI_USERNAME and NAI_PASSWORD). They should be passed via the environment variables (dotenv file supported).\n\n### Examples\nThe examples are in the example folder. Each example is standalone and can be used as a test.\nExamples should be ran with `nox -s run -- python example/<name>.py`.\n\nSome tests can act as example. The full list is as follows :\n- decryption and re-encryption: tests/test_decrypt_encrypt_integrity_check.py\n- diverse generations: tests/test_generate.py\n- parallel generations: tests/test_generate_parallel.py\n\n### Usage\nThe source and all the required functions are located in the novelai-api folder.\nThe examples and tests showcase how this API should be used and can be regarded as the "right way" to use it. However, it doesn\'t mean one can\'t use the "low level" part, which is a thin implementation of the REST endpoints, while the "high level" part is an abstraction built on that low level.\n\n### Contributing\nYou can contribute features and enhancements through PR. Any PR should pass the tests and the pre-commits before submission.\n\nThe tests against the API can be ran with `nox -s test_api`. Note that having node.js installed is required for the test to run properly.\n/!\\ WIP /!\\ The tests against the mocked backend can be ran with `nox -s test_mock`.\n\nTo install and run the pre-commit hook, run `nox -s pre-commit`. This hook should be installed before committing anything.\n',
    'author': 'Aedial',
    'author_email': 'aedial.dev@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Aedial/novelai-api',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.2,<3.12',
}


setup(**setup_kwargs)
